from enum import Enum

from .base import MetricsLayerBase


class MetricsLayerFilterExpressionType(str, Enum):
    Unknown = "UNKNOWN"
    LessThan = "less_than"
    LessOrEqualThan = "less_or_equal_than"
    EqualTo = "equal_to"
    NotEqualTo = "not_equal_to"
    GreaterOrEqualThan = "greater_or_equal_than"
    GreaterThan = "greater_than"
    Like = "like"
    Contains = "contains"
    DoesNotContain = "does_not_contain"
    ContainsCaseInsensitive = "contains_case_insensitive"
    DoesNotContainCaseInsensitive = "does_not_contain_case_insensitive"
    StartsWith = "starts_with"
    EndsWith = "ends_with"
    DoesNotStartWith = "does_not_start_with"
    DoesNotEndWith = "does_not_end_with"
    StartsWithCaseInsensitive = "starts_with_case_insensitive"
    EndsWithCaseInsensitive = "ends_with_case_insensitive"
    DoesNotStartWithCaseInsensitive = "does_not_start_with_case_insensitive"
    DoesNotEndWithCaseInsensitive = "does_not_end_with_case_insensitive"
    IsNull = "is_null"
    IsNotNull = "is_not_null"
    IsIn = "isin"
    IsNotIn = "isnotin"
    BooleanTrue = "boolean_true"
    BooleanFalse = "boolean_false"

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value.lower() == other.value.lower()

    @classmethod
    def parse(cls, value: str):
        try:
            return next(e for e in cls if e.value.lower() == value)
        except StopIteration:
            return cls.Unknown


class Filter(MetricsLayerBase):
    def __init__(self, definition: dict = {}) -> None:
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["field", "value"]
        for k in required_keys:
            if k not in definition:
                raise ValueError(f"Filter missing required key '{k}' The filter passed was {definition}")

    def filter_dict(self):
        return self._filter_dict(self.field, self.value)

    @staticmethod
    def _filter_dict(field: str, value: str):
        # TODO more advanced parsing similar to
        # ref: https://docs.looker.com/reference/field-params/filters

        _symbol_to_filter_type_lookup = {
            "<=": MetricsLayerFilterExpressionType.LessOrEqualThan,
            ">=": MetricsLayerFilterExpressionType.GreaterOrEqualThan,
            "<>": MetricsLayerFilterExpressionType.NotEqualTo,
            "!=": MetricsLayerFilterExpressionType.NotEqualTo,
            "-": MetricsLayerFilterExpressionType.NotEqualTo,
            "=": MetricsLayerFilterExpressionType.EqualTo,
            ">": MetricsLayerFilterExpressionType.GreaterThan,
            "<": MetricsLayerFilterExpressionType.LessThan,
        }

        # Handle null conditiona
        if value == "NULL":
            expression = MetricsLayerFilterExpressionType.IsNull
            cleaned_value = None

        elif value == "-NULL":
            expression = MetricsLayerFilterExpressionType.IsNotNull
            cleaned_value = None

        # isin for strings
        elif len(value.split(", ")) > 1:
            if all(category[0] == "-" for category in value.split(", ")):
                expression = MetricsLayerFilterExpressionType.IsNotIn
                cleaned_value = [f"{category[1:].strip()}" for category in value.split(", ")]

            elif any(category[0] == "-" for category in value.split(", ")):
                raise ValueError("Invalid filter some elements are negated with '-' and some are not")

            else:
                expression = MetricsLayerFilterExpressionType.IsIn
                cleaned_value = [f"{category.strip()}" for category in value.split(", ")]

        # Numeric parsing for less than or equal to, greater than or equal to, not equal to
        elif value[:2] in {"<=", ">=", "<>", "!="}:
            expression = _symbol_to_filter_type_lookup[value[:2]]
            cleaned_value = value[2:]

        # Numeric parsing for equal to, less than, greater than
        elif value[0] in {"=", ">", "<", "-"}:
            expression = _symbol_to_filter_type_lookup[value[0]]
            cleaned_value = value[1:]

        else:
            expression = MetricsLayerFilterExpressionType.EqualTo
            cleaned_value = value

        return {"field": field, "expression": expression, "value": cleaned_value}

    @staticmethod
    def translate_looker_filters_to_sql(sql: str, filters: list):
        case_sql = "case when "
        conditions = []
        for f in filters:
            filter_dict = Filter._filter_dict(f["field"], f["value"])
            parsed_value = filter_dict["value"]

            # Handle null conditiona
            if filter_dict["expression"] == MetricsLayerFilterExpressionType.IsNull:
                condition_value = f"is null"
            elif filter_dict["expression"] == MetricsLayerFilterExpressionType.IsNotNull:
                condition_value = f"is not null"

            # isin or isnotin for strings
            elif filter_dict["expression"] == MetricsLayerFilterExpressionType.IsIn:
                categories = ",".join([f"'{v}'" for v in parsed_value])
                condition_value = f"is in ({categories})"

            elif filter_dict["expression"] == MetricsLayerFilterExpressionType.IsNotIn:
                categories = ",".join([f"'{v}'" for v in parsed_value])
                condition_value = f"is not in ({categories})"

            # Not equal to condition for strings
            elif f["value"][0] == "-":
                condition_value = f"<> '{parsed_value}'"

            # Numeric parsing for less than or equal to, greater than or equal to, not equal to
            elif f["value"][:2] in {"<=", ">=", "<>", "!="}:
                condition_value = f"{f['value'][:2]} {parsed_value}"

            # Numeric parsing for equal to, less than, greater than
            elif f["value"][0] in {"=", ">", "<"}:
                condition_value = f"{f['value'][0]} {parsed_value}"

            else:
                condition_value = f"= '{parsed_value}'"

            field_reference = "${" + f["field"] + "}"
            condition = f"{field_reference} {condition_value}"
            conditions.append(condition)

        # Add the filter conditions AND'd together
        case_sql += " and ".join(conditions)
        # Add the result from the sql arg + imply NULL for anything not hitting the filter condition
        case_sql += f" then {sql} end"

        return case_sql



from copy import deepcopy

from .base import MetricsLayerBase, SQLReplacement
from .field import Field


class Join(MetricsLayerBase, SQLReplacement):
    def __init__(self, definition: dict = {}, explore=None, project=None) -> None:
        self.project = project
        self.explore = explore
        if definition.get("from") is not None:
            definition["from_"] = definition["from"]
        elif definition.get("view_name") is not None:
            definition["from_"] = definition["view_name"]
        else:
            definition["from_"] = definition["name"]

        if "type" not in definition:
            definition["type"] = "left_outer"

        if "relationship" not in definition:
            definition["relationship"] = "many_to_one"

        self.validate(definition)
        super().__init__(definition)

    def replaced_sql_on(self, query_type: str):
        if self.sql_on:
            return self.get_replaced_sql_on(self.sql_on, query_type)
        return f"{self.explore.from_}.{self.foreign_key}={self.from_}.{self.foreign_key}"

    def validate(self, definition: dict):
        required_keys = ["name", "relationship", "type"]

        for k in required_keys:
            if k not in definition:
                join_name = None
                if "name" in definition:
                    join_name = definition["name"]
                raise ValueError(f"Join missing required key {k} in join {join_name}")

        has_sql_on = "sql_on" in definition
        has_fk = "foreign_key" in definition
        has_sql = "sql" in definition
        all_join_arguments = [has_sql_on, has_fk, has_sql]
        no_join_keys = all(not c for c in all_join_arguments)
        multiple_join_keys = sum(all_join_arguments) > 1

        if no_join_keys:
            raise ValueError(
                f"No join arguments found in join {definition['name']}, please pass sql_on or foreign_key"
            )

        if multiple_join_keys:
            raise ValueError(
                f"Multiple join arguments found in join {definition['name']}"
                ", please pass only one of: sql_on, foreign_key"
            )
        super().__init__(definition)

    def collect_errors(self):
        errors = []
        if self.foreign_key:
            for view_name in [self.from_, self.explore.from_]:
                try:
                    self.project.get_field(
                        self.foreign_key, view_name=view_name, explore_name=self.explore.name
                    )
                except Exception:
                    errors.append(
                        f"Could not find field {self.foreign_key} in join {self.name} "
                        f"referencing view {view_name} in explore {self.explore.name}"
                    )
            return errors

        fields_to_replace = self.fields_to_replace(self.sql_on)

        for field in fields_to_replace:
            _, join_name, column_name = Field.field_name_parts(field)
            view_name = self._resolve_view_name(join_name)
            try:
                view = self._get_view_internal(view_name)
            except Exception:
                err_msg = f"Could not find view {view_name} in join {self.name}"
                errors.append(err_msg)
                continue

            try:
                self.project.get_field(column_name, view_name=view.name, explore_name=self.explore.name)
            except Exception:
                errors.append(
                    f"Could not find field {column_name} in join {self.name} "
                    f"referencing view {view_name} in explore {self.explore.name}"
                )
        return errors

    def is_valid(self):
        if self.sql_on:
            fields_to_replace = self.fields_to_replace(self.sql_on)

            # The join isn't valid if we can't find an existing view with that name
            for field in fields_to_replace:
                _, view_name, _ = Field.field_name_parts(field)
                if view_name not in self.explore.join_names():
                    err_msg = (
                        f"Could not find view {view_name} for join {self.name} in explore {self.explore.name}"
                    )
                    print(err_msg)
                    return False
            return True
        is_valid = self.foreign_key is not None
        return is_valid

    def required_views(self):
        if not self.sql_on:
            return [self.explore.from_, self.from_]

        views = []
        for field in self.fields_to_replace(self.sql_on):
            _, join_name, _ = Field.field_name_parts(field)
            if join_name == self.explore.name:
                views.append(self.explore.from_)
            else:
                join = self.explore.get_join(join_name)
                views.append(join.from_)
        return list(set(views))

    def to_dict(self):
        output = {**self._definition}
        return output

    def get_replaced_sql_on(self, sql: str, query_type: str):
        sql_on = deepcopy(sql)
        fields_to_replace = self.fields_to_replace(sql_on)

        for field in fields_to_replace:
            _, join_name, column_name = Field.field_name_parts(field)
            view_name = self._resolve_view_name(join_name)
            view = self._get_view_internal(view_name)

            if view is None:
                return

            table_name = view.name
            field_obj = self.project.get_field(
                column_name, view_name=table_name, explore_name=self.explore.name
            )

            if field_obj and table_name:
                sql_condition = field_obj.sql_query(query_type)
                replace_with = sql_condition
            elif table_name:
                replace_with = f"{table_name}.{column_name}"
            else:
                replace_with = column_name

            replace_text = "${" + field + "}"
            sql_on = sql_on.replace(replace_text, replace_with)

        return sql_on

    def _resolve_view_name(self, join_name: str):
        if join_name == self.explore.name:
            view_name = self.explore.from_
        else:
            join = self.explore.get_join(join_name)
            view_name = join.from_
        return view_name

    def _get_view_internal(self, view_name: str):
        if self.from_ is not None and view_name == self.from_:
            view = self.project.get_view(self.from_)
        elif view_name == self.explore.from_:
            view = self.project.get_view(self.explore.from_)
        else:
            view = self.project.get_view(view_name)
        return view



from typing import List

import networkx

from metrics_layer.core.model.base import MetricsLayerBase
from metrics_layer.core.sql.query_errors import ParseError


class MetricsLayerDesign:
    """ """

    def __init__(self, no_group_by: bool, query_type: str, field_lookup: dict, explore, project) -> None:
        self.no_group_by = no_group_by
        self.query_type = query_type
        self.field_lookup = field_lookup
        self.explore = explore
        self.project = project

    def views(self) -> List[MetricsLayerBase]:
        return self.project.views(explore_name=self.explore.name)

    def joins(self) -> List[MetricsLayerBase]:
        _, access_filter_fields = self.get_access_filter()
        fields_in_query = list(self.field_lookup.values()) + access_filter_fields
        required_views = list(set([v for field in fields_in_query for v in field.required_views()]))
        if self.explore.always_join:
            required_views.extend(self.explore.always_join)

        joins_needed_for_query = []
        for view_name in reversed(sorted(required_views)):
            joins_needed_for_query.extend(self._find_needed_joins(view_name, joins_needed_for_query))
        return self._sort_joins(joins_needed_for_query)

    def _find_needed_joins(self, view_name: str, joins_already_added: list):
        joins_to_add = []

        join_already_added = any(view_name == j.from_ for j in joins_already_added)
        if not join_already_added and view_name != self.explore.from_:
            join = self.explore.get_join(view_name, by_view_name=True)
            if join is None:
                raise ValueError(
                    f"Could not locate join from view {view_name} for explore {self.explore.name}"
                )
            joins_to_add.append(join)
            for view_name in join.required_views():
                joins_to_add.extend(self._find_needed_joins(view_name, joins_already_added + [join]))
        return joins_to_add

    def _sort_joins(self, joins_needed: list):
        if len(joins_needed) == 0:
            return []

        G = networkx.DiGraph()
        for join in joins_needed:
            for view_name in join.required_views():
                G.add_edge(view_name, join.from_)
        ordered_names = list(networkx.bfs_tree(G, source=self.base_view_name))
        # Skip the first one because that's *always* the base of the explore
        return [self.explore.get_join(name, by_view_name=True) for name in ordered_names[1:]]

    def get_view(self, name: str) -> MetricsLayerBase:
        try:
            return next(t for t in self.views() if t.name == name)
        except StopIteration:
            raise ParseError(f"View {name} not found in explore {self.explore.name}")

    def get_join(self, name: str) -> MetricsLayerBase:
        return next((j for j in self.joins() if j.name == name), None)

    def get_field(self, field_name: str, view_name: str = None) -> MetricsLayerBase:
        return self.project.get_field(field_name, view_name=view_name, explore_name=self.explore.name)

    def get_access_filter(self):
        if self.explore.access_filters:
            conditions, fields = [], []
            for condition_set in self.explore.access_filters:
                field = self.project.get_field(condition_set["field"], explore_name=self.explore.name)
                sql = field.sql_query(self.query_type)
                user_attribute_value = condition_set["user_attribute"]

                if self.project._user and self.project._user.get(user_attribute_value):
                    condition = f"{sql} = '{self.project._user[user_attribute_value]}'"
                    conditions.append(condition)
                    fields.append(field)
            return " and ".join(conditions), fields
        return None, []

    @property
    def base_view_name(self):
        return self.explore.from_



import sys

import click

from .seeding import SeedMetricsLayer


def echo(text: str, color: str = None, bold: bool = True):
    if color:
        click.secho(text, fg=color, bold=bold)
    else:
        click.echo(text)


@click.group()
# @click.version_option(version="0.22")
def cli_group():
    pass


@cli_group.command()
def init():
    """Initialize a metrics layer project"""
    SeedMetricsLayer._init_directories()
    SeedMetricsLayer._init_project_file()


@cli_group.command()
@click.option("--connection", default=None, help="The name of the connection to use for the database")
@click.option("--database", help="The name of the database to use for seeding")
@click.option("--schema", default=None, help="The name of the schema to use for seeding")
def seed(connection, database, schema):
    """Seed a metrics layer project by referencing the existing database"""
    SeedMetricsLayer._init_directories()
    profile = SeedMetricsLayer.get_profile()
    seeder = SeedMetricsLayer(profile, connection, database, schema)
    seeder.seed()


@cli_group.command()
def validate():
    """Validate a metrics layer project, internally, without hitting the database"""
    profile = SeedMetricsLayer.get_profile()
    metrics_layer = SeedMetricsLayer._init_profile(profile)
    errors = metrics_layer.config.project.validate()

    if len(errors) == 0:
        n_explores = len(metrics_layer.config.project.explores())
        echo(f"Project passed (checked {n_explores} explore{'s' if n_explores > 1 else ''})!")
    else:
        echo(f"Found {len(errors)} error{'s' if len(errors)> 1 else ''} in the project:\n")
        for error in errors:
            echo(f"\n{error}\n")


@cli_group.command()
def debug():
    """Debug a metrics layer project. Pass your profile name as the sole argument.

    This will list out the inputs and the locations where the metrics layer is finding them,
     in addition to testing the database connection to ensure it works as expected"""
    from metrics_layer import __version__

    profile = SeedMetricsLayer.get_profile()
    metrics_layer = SeedMetricsLayer._init_profile(profile)

    # Environment
    metrics_layer_version = __version__
    echo(f"metrics_layer version: {metrics_layer_version}")

    python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    echo(f"python version: {python_version}")

    python_path = sys.executable
    echo(f"python path: {python_path}")

    profiles_path = metrics_layer.config.profiles_path
    if profiles_path:
        echo(f"Using profiles.yml file at {profiles_path}")
    else:
        echo(f"Could not find profiles.yml file", color="red")

    # Configuration:
    echo(f"\nConfiguration:")
    if metrics_layer.config.profiles_path:
        echo(f"  profiles.yml file OK found and valid", color="green")
    else:
        echo(f"  profiles.yml file Not found", color="red")

    echo(f"\nRequired dependencies:")

    git_status = "OK found" if SeedMetricsLayer._test_git() else "Not found"
    echo(f"  git [{git_status}]", color="green" if "OK" in git_status else "red")

    # Connection:
    connections = metrics_layer.config.connections()
    echo(f"\nConnection{'s' if len(connections) > 1 else ''}:")
    for connection in connections:
        for key, value in connection.printable_attributes().items():
            echo(f"  {key}: {value}")

    test_query = "select 1 as id;"
    for connection in connections:
        try:
            metrics_layer.run_query(test_query, connection)
            connection_status = "OK connection ok"
            color = "green"
        except Exception as e:
            color = "red"
            connection_status = f"Failed with:\n\n {e}"

        echo(f"\nConnection {connection.name} test: {connection_status}", color=color)


@cli_group.command("list")
@click.option(
    "--explore",
    default=None,
    help="The name of the explore (only applicable for fields, dimensions, metrics, and views)",
)
@click.option(
    "--view", default=None, help="The name of the view (only applicable for fields, dimensions, and metrics)"
)
@click.option("--show-hidden", is_flag=True, help="Set this flag if you want to see hidden fields")
@click.argument("type")
def list_(type, explore, view, show_hidden):
    """List attributes in a metrics layer project,
    i.e. models, connections, explores, views, fields, metrics, dimensions"""
    profile = SeedMetricsLayer.get_profile()
    metrics_layer = SeedMetricsLayer._init_profile(profile)

    items = None
    if type == "models":
        items = metrics_layer.list_models(names_only=True)
    elif type == "connections":
        items = metrics_layer.list_connections(names_only=True)
    elif type == "explores":
        items = metrics_layer.list_explores(names_only=True, show_hidden=show_hidden)
    elif type == "views":
        items = metrics_layer.list_views(names_only=True, explore_name=explore, show_hidden=show_hidden)
    elif type == "fields":
        items = metrics_layer.list_fields(
            names_only=True, view_name=view, explore_name=explore, show_hidden=show_hidden
        )
    elif type == "dimensions":
        items = metrics_layer.list_dimensions(
            names_only=True, view_name=view, explore_name=explore, show_hidden=show_hidden
        )
    elif type == "metrics":
        items = metrics_layer.list_metrics(
            names_only=True, view_name=view, explore_name=explore, show_hidden=show_hidden
        )
    elif type == "profiles":
        if profile:
            items = metrics_layer.config.get_all_profiles(names_only=True)
        else:
            from metrics_layer.core.parse import MetricsLayerConfiguration

            default_directory = MetricsLayerConfiguration.get_metrics_layer_directory() + "profiles.yml"
            items = MetricsLayerConfiguration.get_all_profiles(default_directory, names_only=True)
    else:
        click.echo(
            f"Could not find the type {type}, please use one of the options: "
            "models, connections, explores, views, fields, metrics, dimensions"
        )
    if items:
        click.echo(f"Found {len(items)} {type if len(items) > 1 else type[:-1]}:\n")
        for name in items:
            click.echo(name)

    if items is not None and len(items) == 0:
        click.echo(f"Could not find any {type}")


@cli_group.command()
@click.option(
    "--type",
    help="The type of object to show. One of: model, connection, explore, view, field, metric, dimension",  # noqa
)
@click.option(
    "--explore",
    default=None,
    help="The name of the explore (only applicable for fields, dimensions, metrics, and views)",
)
@click.option(
    "--view", default=None, help="The name of the view (only applicable for fields, dimensions, and metrics)"
)
@click.argument("name")
def show(type, name, explore, view):
    """Show information on an attribute in a metrics layer project, by name"""
    profile = SeedMetricsLayer.get_profile()
    metrics_layer = SeedMetricsLayer._init_profile(profile)

    attributes = []
    if type == "model":
        attributes = metrics_layer.get_model(name)
    elif type == "connection":
        attributes = metrics_layer.get_connection(name)
    elif type == "explore":
        attributes = metrics_layer.get_explore(name)
    elif type == "view":
        attributes = metrics_layer.get_view(name, explore_name=explore)
    elif type == "field":
        attributes = metrics_layer.get_field(name, view_name=view, explore_name=explore)
    elif type == "dimension":
        attributes = metrics_layer.get_dimension(name, view_name=view, explore_name=explore)
    elif type == "metric":
        attributes = metrics_layer.get_metric(name, view_name=view, explore_name=explore)
    else:
        click.echo(
            f"Could not find the type {type}, please use one of the options: "
            "models, connections, explores, views, fields, metrics, dimensions"
        )

    if attributes:
        click.echo(f"Attributes in {type} {name}:\n")
        for key, value in attributes.printable_attributes().items():
            if isinstance(value, list) and len(value) > 0:
                click.echo(f"  {key}:")
                for item in value:
                    click.echo(f"    {item}")
            else:
                click.echo(f"  {key}: {value}")


if __name__ == "__main__":
    cli_group()



import json
import os

import pandas as pd


class SeedMetricsLayer:
    def __init__(self, profile, connection=None, database=None, schema=None):
        self.profile_name = profile
        self.metrics_layer, self.connection = self._init_connection(self.profile_name, connection)
        self.database = database if database else self.connection.database
        self.schema = schema if schema else self.connection.schema
        if schema:
            self.location_description = f"schema: {self.database}.{self.schema}"
        else:
            self.location_description = f"database: {self.database}"

        self._snowflake_type_lookup = {
            "DATE": "date",
            "DATETIME": "timestamp",
            "TIMESTAMP_TZ": "timestamp",
            "TIMESTAMP_NTZ": "timestamp",
            "TIMESTAMP_LTZ": "timestamp",
            "TIMESTAMP": "timestamp",
            "BOOLEAN": "yesno",
            "FIXED": "number",
            "FLOAT": "number",
            "REAL": "number",
            "TEXT": "string",
            "VARCHAR": "string",
            "CHAR": "string",
            "CHARACTER": "string",
            "STRING": "string",
            "TEXT": "string",
            "BINARY": "string",
            "VARBINARY": "string",
        }

    def seed(self):
        from metrics_layer.core.model.definitions import Definitions
        from metrics_layer.core.parse.project_reader import ProjectReader

        if self.connection.type != Definitions.snowflake:
            raise NotImplementedError(
                "The only data warehouse supported for seeding at this time is Snowflake"
            )
        table_query = self.table_query(type="tables")
        table_data = self.run_query(table_query)
        table_df = self.table_result_to_dataframe(table_data)
        print(f"Got information on all tables in {self.location_description}")

        view_query = self.table_query(type="views")
        view_data = self.run_query(view_query)
        view_df = self.view_result_to_dataframe(view_data)
        print(f"Got information on all views in {self.location_description}")

        data = self.join_view_and_table_results(table_df, view_df)

        # Each row represents either a single table or a single view
        views = []
        for i, (_, row) in enumerate(data.iterrows()):
            columns_query = self.columns_query(row["NAME"], row["SCHEMA_NAME"], row["DATABASE_NAME"])
            column_data = self.run_query(columns_query)
            column_df = self.column_result_to_dataframe(column_data)

            column_df["RAW_TYPE"] = column_df["DATA_TYPE"].apply(lambda x: json.loads(x).get("type", "TEXT"))

            view = self.make_view(column_df, row["NAME"], row["SCHEMA_NAME"])
            print(f"Got information on {row['TYPE']} {row['NAME']} ({i + 1} / {len(data)})")
            views.append(view)

        models = self.make_models(views)

        reader = ProjectReader(None)
        # Fake that the project was loaded from a repo
        reader.unloaded = False

        reader._models = models
        reader._views = views
        # Dump the models to yaml files
        reader.dump(os.getcwd())

    def make_models(self, views: list):
        model = {
            "version": 1,
            "type": "model",
            "name": "base_model",
            "connection": self.connection.name,
            "explores": [{"name": view["name"]} for view in views],
        }
        return [model]

    def make_view(self, column_data, table_name: str, schema_name: str):
        view_name = self.clean_name(table_name)
        count_measure = {"field_type": "measure", "name": "count", "type": "count"}
        fields = self.make_fields(column_data) + [count_measure]
        view = {
            "version": 1,
            "type": "view",
            "name": view_name,
            "sql_table_name": f"{schema_name}.{table_name}",
            "default_date": next((f["name"] for f in fields if f["field_type"] == "dimension_group"), None),
            "row_label": "TODO - Label row",
            "fields": fields,
        }
        return view

    def make_fields(self, column_data: str):
        fields = []
        for _, row in column_data[["COLUMN_NAME", "RAW_TYPE"]].iterrows():
            name = self.clean_name(row["COLUMN_NAME"])
            metrics_layer_type = self._snowflake_type_lookup.get(row["RAW_TYPE"], "string")
            sql = "${TABLE}." + row["COLUMN_NAME"]

            field = {"name": name, "sql": sql}

            if metrics_layer_type in {"timestamp", "date"}:
                field["field_type"] = "dimension_group"
                field["type"] = "time"
                field["timeframes"] = ["raw", "date", "week", "month", "quarter", "year"]
                field["datatype"] = metrics_layer_type
            else:
                field["field_type"] = "dimension"
                field["type"] = metrics_layer_type
            fields.append(field)
        return fields

    def table_query(self, type="tables"):
        query = f"show {type} in "
        if self.database and not self.schema:
            query += f"database {self.database};"
        elif self.database and self.schema:
            query += f"schema {self.database}.{self.schema};"
        else:
            raise ValueError("You must specify a database or a database and a schema for seeding")
        return query

    @staticmethod
    def table_result_to_dataframe(table_data):
        table_df = pd.DataFrame(
            [{"NAME": i[1], "DATABASE_NAME": i[2], "SCHEMA_NAME": i[3], "TYPE": "table"} for i in table_data]
        )
        return table_df

    @staticmethod
    def view_result_to_dataframe(view_data):
        view_df = pd.DataFrame(
            [{"NAME": i[1], "DATABASE_NAME": i[3], "SCHEMA_NAME": i[4], "TYPE": "view"} for i in view_data]
        )
        return view_df

    @staticmethod
    def join_view_and_table_results(table_df, view_df):
        columns = ["DATABASE_NAME", "SCHEMA_NAME", "NAME", "TYPE"]
        if view_df.empty and table_df.empty:
            print("Could not find any tables or views, are you sure you are looking ar the right database?")
            raise ValueError("No tables or views found")
        if view_df.empty:
            return table_df[columns]
        if table_df.empty:
            return view_df[columns]
        return pd.concat([table_df[columns], view_df[columns]], sort=False)

    def columns_query(self, table_name: str, schema_name: str, database_name: str):
        return f'show columns in "{database_name}"."{schema_name}"."{table_name}";'

    @staticmethod
    def column_result_to_dataframe(column_data):
        column_df = pd.DataFrame([{"COLUMN_NAME": i[2], "DATA_TYPE": i[3]} for i in column_data])
        return column_df

    def run_query(self, query: str):
        return self.metrics_layer.run_query(query, self.connection, raw_cursor=True)

    @staticmethod
    def _init_connection(profile_name: str, connection_name: str = None):
        metrics_layer = SeedMetricsLayer._init_profile(profile_name)

        if connection_name:
            connection = metrics_layer.config.get_connection(connection_name)
        else:
            connections = metrics_layer.config.connections()
            if len(connections) == 1:
                connection = connections[0]
            else:
                raise ValueError(
                    f"Could not determine the connection to use with profile {profile_name}, "
                    "please pass the connection name with the --connection arg"
                )
        return metrics_layer, connection

    @staticmethod
    def _init_profile(profile_name: str):
        from metrics_layer.core import MetricsLayerConnection

        metrics_layer = MetricsLayerConnection(profile_name)
        return metrics_layer

    @staticmethod
    def get_profile():
        from metrics_layer.core.parse.project_reader import ProjectReader

        zenlytic_project_path = os.path.join(os.getcwd(), "zenlytic_project.yml")
        dbt_project_path = os.path.join(os.getcwd(), "dbt_project.yml")

        if os.path.exists(zenlytic_project_path):
            zenlytic_project = ProjectReader.read_yaml_file(zenlytic_project_path)
            return zenlytic_project["profile"]
        elif os.path.exists(dbt_project_path):
            dbt_project = ProjectReader.read_yaml_file(dbt_project_path)
            return dbt_project["profile"]
        raise ValueError(
            """Could not find a profile for the metrics layer in either the zenlytic_project.yml file or in
         the dbt_project.yml file, if neither of those files exist, please create the file"""
        )

    @staticmethod
    def _init_directories():
        models_dir = "models/"
        views_dir = "views/"
        for directory in [models_dir, views_dir]:
            fully_qualified_path = os.path.join(os.getcwd(), directory)
            if not os.path.exists(fully_qualified_path):
                os.mkdir(fully_qualified_path)

    @staticmethod
    def _init_project_file():
        from metrics_layer.core.parse.project_reader import ProjectReader

        default_profile = {"name": "zenlytic_project_name", "profile": "my_dbt_profile", "folder": "./"}
        ProjectReader._dump_yaml_file(default_profile, "zenlytic_project.yml")

    @staticmethod
    def _test_git():
        try:
            import git  # noqa

            valid = True
        except ImportError:
            valid = False
        return valid

    @staticmethod
    def clean_name(txt: str):
        return txt.lower().replace(" ", "_")



from .base import AccessDeniedOrDoesNotExistException
from .dashboard import Dashboard
from .explore import Explore
from .field import Field
from .model import AccessGrant, Model
from .view import View


class Project:
    """
    Higher level abstraction for the whole project
    """

    def __init__(
        self,
        models: list,
        views: list,
        dashboards: list = [],
        looker_env: str = None,
        connection_lookup: dict = {},
        manifest=None,
    ):
        self._models = models
        self._views = views
        self._dashboards = dashboards
        self.looker_env = looker_env
        self.connection_lookup = connection_lookup
        self.manifest = manifest
        self.manifest_exists = manifest and manifest.exists()
        self._user = None

    def __repr__(self):
        text = "models" if len(self._models) != 1 else "model"
        return f"<Project {len(self._models)} {text}>"

    def set_user(self, user: dict):
        self._user = user

    def validate(self):
        all_errors = []
        for explore in self.explores():
            errors = explore.validate_fields()
            all_errors.extend(errors)

        for dashboard in self.dashboards():
            errors = dashboard.collect_errors()
            all_errors.extend(errors)
        return all_errors

    def _all_dashboards(self):
        dashboards = []
        for d in self._dashboards:
            dashboard = Dashboard(d, project=self)
            user_allowed = self.can_access_dashboard(dashboard)
            if user_allowed:
                dashboards.append(dashboard)
        return dashboards

    def dashboards(self) -> list:
        return self._all_dashboards()

    def get_dashboard(self, dashboard_name: str) -> Model:
        try:
            return next((d for d in self.dashboards() if d.name == dashboard_name))
        except StopIteration:
            raise AccessDeniedOrDoesNotExistException(
                f"Could not find or you do not have access to dashboard {dashboard_name}",
                object_name=dashboard_name,
                object_type="dashboard",
            )

    def models(self) -> list:
        return [Model(m) for m in self._models]

    def get_model(self, model_name: str) -> Model:
        return next((m for m in self.models() if m.name == model_name), None)

    def access_grants(self):
        return [AccessGrant(g) for m in self.models() for g in m.access_grants]

    def get_access_grant(self, grant_name: str):
        return next((ag for ag in self.access_grants() if ag.name == grant_name), None)

    def can_access_dashboard(self, dashboard: Dashboard):
        return self._can_access_object(dashboard)

    def can_access_explore(self, explore: Explore):
        return self._can_access_object(explore)

    def can_access_join(self, join, explore: Explore):
        can_access_explore = self.can_access_explore(explore)
        return self._can_access_object(join) and can_access_explore

    def can_access_view(self, view: View):
        return self._can_access_object(view)

    def can_access_field(self, field):
        can_access_view = self.can_access_view(field.view)
        return self._can_access_object(field) and can_access_view

    def _can_access_object(self, obj):
        if self._user is not None:
            if obj.required_access_grants:
                decisions = []
                for grant_name in obj.required_access_grants:
                    grant = self.get_access_grant(grant_name)
                    user_attribute_value = self._user.get(grant.user_attribute)

                    if user_attribute_value is None:
                        decision = True
                    else:
                        decision = user_attribute_value in grant.allowed_values
                    decisions.append(decision)

                # We use all here because the condition between access conditions is AND
                return all(decisions)
        return True

    def _all_explores(self):
        explores = []
        for m in self.models():
            for e in m.explores:
                explore = Explore({**e, "model": m}, project=self)
                user_allowed = self.can_access_explore(explore)
                if user_allowed:
                    explores.append(explore)
        return explores

    def explores(self, show_hidden: bool = True) -> list:
        explores = self._all_explores()
        if show_hidden:
            return explores
        return [e for e in explores if e.hidden == "no" or not e.hidden]

    def get_explore(self, explore_name: str) -> Explore:
        try:
            return next((e for e in self.explores() if e.name == explore_name))
        except StopIteration:
            raise AccessDeniedOrDoesNotExistException(
                f"Could not find or you do not have access to explore {explore_name}",
                object_name=explore_name,
                object_type="explore",
            )

    def _all_views(self, explore):
        views = []
        for v in self._views:
            view = View({**v, "explore": explore}, project=self)
            if self.can_access_view(view):
                views.append(view)
        return views

    def views(self, explore_name: str = None, explore: Explore = None) -> list:
        if explore_name:
            return self.views_with_explore(explore_name=explore_name)
        return self._all_views(explore)

    def views_with_explore(self, explore_name: str = None, explore: Explore = None):
        if explore_name and explore is None:
            explore = self.get_explore(explore_name)
        view_names = [explore.from_] + [j.from_ for j in explore.joins()]
        return [v for v in self.views(explore=explore) if v.name in view_names]

    def get_view(self, view_name: str, explore: Explore = None) -> View:
        try:
            return next((v for v in self.views(explore=explore) if v.name == view_name))
        except StopIteration:
            raise AccessDeniedOrDoesNotExistException(
                f"Could not find or you do not have access to view {view_name}",
                object_name=view_name,
                object_type="view",
            )

    def sets(self, view_name: str = None, explore_name: str = None):
        if explore_name:
            views = self.views(explore_name=explore_name)
        elif view_name:
            try:
                views = [self.get_view(view_name)]
            except AccessDeniedOrDoesNotExistException:
                views = []
        else:
            views = self.views()

        all_sets = []
        for view in views:
            all_sets.extend(view.list_sets())
        return all_sets

    def get_set(self, set_name: str, view_name: str = None):
        if view_name:
            sets = self.sets(view_name=view_name)
        else:
            sets = self.sets()
        return next((s for s in sets if s.name == set_name), None)

    def fields(
        self,
        explore_name: str = None,
        view_name: str = None,
        show_hidden: bool = True,
        expand_dimension_groups: bool = False,
        show_excluded: bool = False,
    ) -> list:
        if explore_name is None and view_name is None:
            return self._all_fields(show_hidden, expand_dimension_groups)
        elif view_name and explore_name:
            return self._view_fields(
                view_name, self.get_explore(explore_name), show_hidden, expand_dimension_groups
            )
        elif view_name:
            return self._view_fields(
                view_name, show_hidden=show_hidden, expand_dimension_groups=expand_dimension_groups
            )
        else:
            return self._explore_fields(explore_name, show_hidden, expand_dimension_groups, show_excluded)

    def _all_fields(self, show_hidden: bool, expand_dimension_groups: bool):
        return [f for v in self.views() for f in v.fields(show_hidden, expand_dimension_groups)]

    def _view_fields(
        self,
        view_name: str,
        explore: Explore = None,
        show_hidden: bool = True,
        expand_dimension_groups: bool = False,
    ):
        view = self.get_view(view_name, explore=explore)
        if not view:
            plus_explore = f" in explore {explore.name}" if explore else ""
            raise ValueError(f"Could not find a view matching the name {view_name}{plus_explore}")
        return view.fields(show_hidden, expand_dimension_groups)

    def _explore_fields(
        self, explore_name: str, show_hidden: bool, expand_dimension_groups: bool, show_excluded: bool
    ):
        explore = self.get_explore(explore_name)
        valid_views = self.views_with_explore(explore=explore)
        fields = [f for v in valid_views for f in v.fields(show_hidden, expand_dimension_groups)]

        if explore.fields and not show_excluded:
            field_names = explore.field_names()
            fields_in_explore = []
            for field in fields:
                if any(field.equal(fn) for fn in field_names):
                    fields_in_explore.append(field)
            return fields_in_explore
        return fields

    def get_field(self, field_name: str, explore_name: str = None, view_name: str = None) -> Field:
        # Handle the case where the explore syntax is passed: explore_name.field_name
        if "." in field_name:
            specified_explore_name, specified_view_name, field_name = Field.field_name_parts(field_name)
            if view_name and specified_view_name != view_name:
                raise ValueError(
                    f"You specificed two different view names {specified_view_name} and {view_name}"
                )
            if specified_explore_name and explore_name and specified_explore_name != explore_name:
                raise ValueError(
                    f"You specificed two different explore names {specified_explore_name} and {explore_name}"
                )
            view_name = specified_view_name
            if specified_explore_name:
                explore_name = specified_explore_name

        field_name = field_name.lower()
        fields = self.fields(explore_name=explore_name, view_name=view_name)
        matching_fields = [f for f in fields if f.equal(field_name)]
        return self._matching_field_handler(matching_fields, field_name, explore_name, view_name)

    def get_explore_from_field(self, field_name: str):
        # If it's specified this is really easy
        if "." in field_name:
            explore_name, _, _ = Field.field_name_parts(field_name)
            if explore_name is not None:
                return explore_name

        # If it's not we have to check all explores to make sure the field isn't ambiguously referenced
        all_fields_with_explore_duplicates = []
        for explore in self.explores():
            for view in self.views_with_explore(explore_name=explore.name):
                all_fields_with_explore_duplicates.extend(view.fields())

        matching_fields = [f for f in all_fields_with_explore_duplicates if f.name == field_name]
        match = self._matching_field_handler(matching_fields, field_name)
        return match.view.explore.name

    def _matching_field_handler(
        self, matching_fields: list, field_name: str, explore_name: str = None, view_name: str = None
    ):
        if len(matching_fields) == 1:
            return matching_fields[0]

        elif len(matching_fields) > 1:
            matching_names = [self._fully_qualified_name(f) for f in matching_fields]
            explore_text = f", in explore {explore_name}" if explore_name else ""
            view_text = f", in view {view_name}" if view_name else ""
            err_msg = (
                f"Multiple fields found for the name {field_name}{explore_text}{view_text}"
                f" - those fields were {matching_names} \n\nPlease specify a "
                "view name like this: 'view_name.field_name' or "
                "an explore and view like this 'explore_name.view_name.field_name'"
                "\n\nor pass the argument 'explore_name' to the function, to set the explore"
            )
            raise ValueError(err_msg)
        elif field_name == "count" and (explore_name or view_name):
            definition = {"type": "count", "name": "count", "field_type": "measure"}
            if explore_name and not view_name:
                view_name = self.get_explore(explore_name).from_
            return Field(definition, view=self.get_view(view_name))
        else:
            err_msg = f"Field {field_name} not found"
            if explore_name:
                err_msg += f" in explore {explore_name}"
            if view_name:
                err_msg += f" in view {view_name}"
            err_msg += ", please check that this field exists AND that you have access to it. \n\n"
            err_msg += "If this is a dimension group specify the group parameter, if not already specified, "
            err_msg += "for example, with a dimension group named 'order' with timeframes: [raw, date, month]"
            err_msg += " specify 'order_raw' or 'order_date' or 'order_month'"
            raise AccessDeniedOrDoesNotExistException(err_msg, object_name=field_name, object_type="field")

    def resolve_dbt_ref(self, ref_name: str, view_name: str = None):
        if not self.manifest_exists:
            raise ValueError(
                f"Could not find a dbt project co-located with this "
                f"project to resolve the dbt ref('{ref_name}') in view {view_name}"
            )
        return self.manifest.resolve_name(ref_name)

    @staticmethod
    def _fully_qualified_name(field: Field):
        name = f"{field.view.name}.{field.name}"
        if field.view.explore:
            name = f"{field.view.explore.name}.{name}"
        return name



from .base import MetricsLayerBase


class AccessGrant(MetricsLayerBase):
    def __init__(self, definition: dict = {}) -> None:

        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "user_attribute", "allowed_values"]
        for k in required_keys:
            if k not in definition:
                raise ValueError(f"Access Grant missing required key {k}")


class Model(MetricsLayerBase):
    def __init__(self, definition: dict = {}) -> None:

        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "connection"]
        for k in required_keys:
            if k not in definition:
                raise ValueError(f"Model missing required key {k}")

    def printable_attributes(self):
        to_print = ["name", "type", "label", "group_label", "connection", "explore_names"]
        attributes = self.to_dict()
        attributes["explore_names"] = [e["name"] for e in attributes["explores"]]
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    @property
    def explores(self):
        return self._definition.get("explores", [])

    @property
    def access_grants(self):
        return self._definition.get("access_grants", [])



import json
import os
from copy import deepcopy

import lkml
import yaml

from metrics_layer.core.utils import merge_nested_dict

from .github_repo import BaseRepo


class ProjectReader:
    def __init__(self, repo: BaseRepo, additional_repo: BaseRepo = None, profiles_dir: str = None):
        self.base_repo = repo
        self.additional_repo = additional_repo
        self.multiple_repos = self.additional_repo is not None
        self.profiles_dir = profiles_dir
        self.version = 1
        self.unloaded = True
        self.has_dbt_project = False
        self.manifest = {}
        self._models = []
        self._views = []
        self._dashboards = []

    @property
    def models(self):
        if self.unloaded:
            self.load()
        return self._models

    @property
    def views(self):
        if self.unloaded:
            self.load()
        return self._views

    @property
    def dashboards(self):
        if self.unloaded:
            self.load()
        return self._dashboards

    def dump(self, path: str):
        for model in self.models:
            file_name = model["name"] + "_model.yml"
            models_folder = os.path.join(path, "models/")
            if os.path.exists(models_folder):
                file_path = os.path.join(models_folder, file_name)
            else:
                file_path = os.path.join(path, file_name)
            self._dump_yaml_file(model, file_path)

        for view in self.views:
            file_name = view["name"] + "_view.yml"
            views_folder = os.path.join(path, "views/")
            if os.path.exists(views_folder):
                file_path = os.path.join(views_folder, file_name)
            else:
                file_path = os.path.join(path, file_name)
            self._dump_yaml_file(view, file_path)

    @staticmethod
    def _dump_yaml_file(data: dict, path: str):
        with open(path, "w") as f:
            yaml.dump(data, f)

    def load(self) -> None:
        base_models, base_views, base_dashboards = self._load_repo(self.base_repo)
        if self.multiple_repos:
            additional_models, additional_views, additional_dashboards = self._load_repo(self.additional_repo)
            self._models = self._merge_objects(base_models, additional_models)
            self._views = self._merge_objects(base_views, additional_views)
            self._dashboards = base_dashboards + additional_dashboards
        else:
            self._models = base_models
            self._views = base_views
            self._dashboards = base_dashboards

        self.unloaded = False

    def _load_repo(self, repo: BaseRepo):
        repo.fetch()
        repo_type = repo.get_repo_type()
        if repo_type == "lookml":
            models, views, dashboards = self._load_lookml(repo)
        elif repo_type == "dbt":
            models, views, dashboards = self._load_dbt(repo)
        elif repo_type == "metrics_layer":
            models, views, dashboards = self._load_metrics_layer(repo)
        else:
            raise TypeError(
                f"Unknown repo type: {repo_type}, valid values are 'metrics_layer', 'lookml', 'dbt'"
            )
        repo.delete()
        return models, views, dashboards

    def _load_lookml(self, repo: BaseRepo):
        models = []
        for fn in repo.search(pattern="*.model.*"):
            model_name = self._parse_model_name(fn)
            model = self.read_lkml_file(fn)
            models.append(self._standardize_model({**model, "name": model_name, "type": "model"}))

        views = []
        for fn in repo.search(pattern="*.view.*"):
            file_views = self.read_lkml_file(fn).get("views", [])
            views.extend([self._standardize_view(v) for v in file_views])

        # Empty list is for currently unsupported dashboards when using the lookml mode
        return models, views, []

    def _load_dbt(self, repo: BaseRepo):
        self.project_name = self._get_dbt_project_file(repo.folder)["name"]
        self._dump_profiles_file(repo.folder, self.project_name)
        self._generate_manifest_json(repo.folder, self.profiles_dir)

        self.manifest = self._load_manifest_json(repo)
        models, views = self._parse_dbt_manifest(self.manifest)

        # Empty list is for currently unsupported dashboards when using the dbt mode
        return models, views, []

    @staticmethod
    def _load_manifest_json(repo):
        manifest_files = repo.search(pattern="manifest.json")
        if len(manifest_files) > 1:
            raise ValueError("found multiple manifest.json files for your dbt project")
        if len(manifest_files) == 0:
            raise ValueError("could not find a manifest.json file for your dbt project")

        with open(manifest_files[0], "r") as f:
            manifest = json.load(f)
        return manifest

    def _parse_dbt_manifest(self, manifest: dict):
        views = self._make_dbt_views(manifest)
        models = self._make_dbt_models([v["name"] for v in views])
        return models, views

    def _make_dbt_views(self, manifest: dict):
        metrics = [self._make_dbt_metric(m) for m in manifest["metrics"].values()]
        view_keys = [k for k in manifest["nodes"].keys() if "model." in k]

        views = []
        for view_key in view_keys:
            view_raw = manifest["nodes"][view_key]
            view_metrics = [m for m in metrics if view_raw["name"] == m.get("model")]
            if len(view_metrics) > 0:
                view = self._make_dbt_view(view_raw, view_metrics)
                views.append(view)

        return views

    def _make_dbt_view(self, view: dict, view_metrics: list):
        unique_timestamps = list({m["timestamp"] for m in view_metrics})
        # Pick the longest time grain
        time_grains = list(sorted(view_metrics, key=lambda x: len(x["time_grains"])))[-1]

        if len(unique_timestamps) > 1:
            raise ValueError(
                "cannot handle dbt metrics with different primary timestamps in a view / dbt model"
            )
        dimension_group = self._make_dbt_dimension_group(unique_timestamps[0], time_grains)
        metrics = []
        for m in view_metrics:
            m.pop("model", None)
            m.pop("timestamp", None)
            m.pop("time_grains", None)
            metrics.append(m)

        extra = view["meta"] if view["meta"] != {} else view.get("config", {}).get("meta", {})
        dimensions = [
            self._make_dbt_dimension(d) for d in view.get("columns").values() if d.get("is_dimension")
        ]
        default_date = unique_timestamps[0]
        fields = dimensions + [dimension_group] + metrics
        view_dict = {
            "version": 1,
            "type": "view",
            "name": view["name"],
            "description": view.get("description"),
            "row_label": extra.get("row_label"),
            "sql_table_name": f"{view['schema']}.{view['name']}",
            "default_date": default_date,
            "extra": extra,
            "fields": fields,
        }
        return view_dict

    @staticmethod
    def _make_dbt_dimension_group(name: str, time_grains: list):
        return {
            "field_type": "dimension_group",
            "name": name,
            "type": "time",
            "datatype": "date",
            "timeframes": [t if t != "day" else "date" for t in time_grains],
            "sql": "${TABLE}." + name,
        }

    @staticmethod
    def _make_dbt_dimension(dimension: dict):
        return {
            "field_type": "dimension",
            "name": dimension["name"],
            "type": "string",
            "label": dimension.get("label"),
            "sql": "${TABLE}." + dimension["name"],
            "description": dimension.get("description"),
            "extra": dimension.get("meta", {}),
            "value_format_name": dimension.get("meta", {}).get("value_format_name"),
        }

    @staticmethod
    def _make_dbt_metric(metric: dict):
        if len(metric["sql"].split(" ")) > 1:
            raise ValueError(
                "We do not currently support dbt sql statements that are more than an identifier"
            )
        if any(f["operator"] != "equal_to" for f in metric.get("filters", [])):
            raise ValueError("We do not currently support dbt filter statements that are not equal_to")
        metric_dict = {
            "name": metric["name"],
            "model": metric["model"].replace("ref('", "").replace("')", ""),
            "timestamp": metric["timestamp"],
            "time_grains": metric["time_grains"],
            "field_type": "measure",
            "type": metric["type"],
            "label": metric.get("label"),
            "description": metric.get("description"),
            "sql": "${TABLE}." + metric["sql"],
            "extra": metric.get("meta", {}),
            "filters": [{"field": f["field"], "value": f["value"]} for f in metric.get("filters", [])],
        }
        return metric_dict

    def _make_dbt_models(self, view_names: list):
        model = {"version": 1, "type": "model", "name": self.project_name, "connection": self.project_name}
        model["explores"] = [{"name": view_name} for view_name in view_names]
        return [model]

    def _get_dbt_project_file(self, project_dir: str):
        dbt_project = self.read_yaml_file(os.path.join(project_dir, "dbt_project.yml"))
        return dbt_project

    def _dump_profiles_file(self, project_dir: str, project_name: str):
        # It doesn't matter the warehouse type here because we're just compiling the models
        params = {
            "type": "snowflake",
            "account": "fake-url.us-east-1",
            "user": "fake",
            "password": "fake",
            "warehouse": "fake",
            "database": "fake",
            "schema": "fake",
        }

        profiles = {
            project_name: {"target": "temp", "outputs": {"temp": {**params}}},
            "config": {"send_anonymous_usage_stats": False},
        }
        self._dump_yaml_file(profiles, os.path.join(project_dir, "profiles.yml"))

    def _generate_manifest_json(self, project_dir: str, profiles_dir: str):

        if profiles_dir is None:
            profiles_dir = project_dir
            if not os.path.exists(os.path.join(profiles_dir, "profiles.yml")):
                project = self._get_dbt_project_file(project_dir)
                self._dump_profiles_file(profiles_dir, project["profile"])

        self._run_dbt("ls", project_dir=project_dir, profiles_dir=profiles_dir)

    @staticmethod
    def _run_dbt(cmd: str, project_dir: str, profiles_dir: str):
        # from dbt.main import handle_and_check

        os.system(f"dbt {cmd} --project-dir {project_dir} --profiles-dir {profiles_dir}")

    def _load_metrics_layer(self, repo: BaseRepo):
        models, views, dashboards = [], [], []
        self.has_dbt_project = len(list(repo.search(pattern="dbt_project.yml"))) > 0
        if self.has_dbt_project:
            self._generate_manifest_json(repo.folder, self.profiles_dir)
            self.manifest = self._load_manifest_json(repo)

        file_names = repo.search(pattern="*.yml") + repo.search(pattern="*.yaml")
        for fn in file_names:
            yaml_dict = self.read_yaml_file(fn)

            # Handle keyerror
            if "type" not in yaml_dict:
                print(f"WARN: file {fn} is missing a type")

            yaml_type = yaml_dict.get("type")

            if yaml_type == "model":
                models.append(yaml_dict)
            elif yaml_type == "view":
                views.append(yaml_dict)
            elif yaml_type == "dashboard":
                dashboards.append(yaml_dict)
            elif yaml_type:
                print(f"WARN: Unknown file type '{yaml_type}' options are 'model', 'view', or 'dashboard'")

        return models, views, dashboards

    def _standardize_view(self, view: dict):
        # Get all fields under the same key "fields"
        measures = self._fields_with_type(view.pop("measures", []), "measure")
        dimensions = self._fields_with_type(view.pop("dimensions", []), "dimension")
        dimension_groups = self._fields_with_type(view.pop("dimension_groups", []), "dimension_group")

        fields = [self._standardize_field(f) for f in measures + dimensions + dimension_groups]
        return {**view, "type": "view", "fields": fields}

    def _standardize_model(self, model: dict):
        model["explores"] = [self._standardize_explore(e) for e in model.get("explores", [])]
        return model

    def _standardize_explore(self, explore: dict):
        if "always_filter" in explore:
            filters = explore["always_filter"].pop("filters__all")
            explore["always_filter"]["filters"] = self._standardize_filters(filters)
        return explore

    def _standardize_field(self, field: dict):
        if "filters__all" in field:
            filters = field.pop("filters__all")
            field["filters"] = self._standardize_filters(filters)
        return field

    def _standardize_filters(self, filters: list):
        clean_filters = []
        for f in filters:
            if isinstance(f, list):
                for nested_filter in f:
                    clean_filters.append(self.__clean_filter(nested_filter))
            else:
                clean_filters.append(self.__clean_filter(f))
        return clean_filters

    def __clean_filter(self, filter_dict: dict):
        # OG looker filter pattern
        if "field" in filter_dict and "value" in filter_dict:
            return filter_dict

        # New Looker filter pattern
        field = list(filter_dict.keys())[0]
        value = list(filter_dict.values())[0]
        return {"field": field, "value": value}

    def _merge_objects(self, base_objects: list, additional_objects: list):
        results = []
        for obj in base_objects:
            additional_obj = self._get_from_list(obj["name"], additional_objects)
            if additional_obj:
                final_obj = merge_nested_dict(base=deepcopy(obj), additional=deepcopy(additional_obj))
            else:
                final_obj = obj
            results.append(final_obj)
        return results

    @staticmethod
    def _get_from_list(name: str, objects: list):
        return next((i for i in objects if i["name"] == name), None)

    @staticmethod
    def read_lkml_file(path: str):
        with open(path, "r") as file:
            lkml_dict = lkml.load(file)
        return lkml_dict

    @staticmethod
    def _fields_with_type(fields: list, field_type: str):
        return [{**f, "field_type": field_type} for f in fields]

    @staticmethod
    def _parse_model_name(model_path: str) -> str:
        return model_path.split("/")[-1].replace(".model.", "").replace("lkml", "").replace("lookml", "")

    @staticmethod
    def read_yaml_file(path: str):
        with open(path, "r") as file:
            yaml_dict = yaml.safe_load(file)
        return yaml_dict



from pypika import Criterion


class LiteralValueCriterion(Criterion):
    def __init__(self, sql_query: str, alias: str = None) -> None:
        """A wrapper for a literal value criterion which is a string of valid sql"""
        super().__init__(alias)
        self.sql_query = sql_query

    def get_sql(self, **kwargs):
        return self.sql_query



class Manifest:
    def __init__(self, definition: dict):
        self._definition = definition

    def exists(self):
        return self._definition is not None and self._definition != {}

    def _resolve_node(self, name: str):
        key = next((k for k in self._definition["nodes"].keys() if name == k.split(".")[-1]), None)
        if key is None:
            raise ValueError(
                f"Could not find the ref {name} in the co-located dbt project."
                " Please check the name in your dbt project."
            )
        return self._definition["nodes"][key]

    def resolve_name(self, name: str):
        node = self._resolve_node(name)
        # return f"{node['database']}.{node['schema']}.{node['alias']}"
        return f"{node['schema']}.{node['alias']}"



import datetime
from copy import deepcopy
from typing import Dict

import sqlparse
from pypika import Criterion, Field
from pypika.terms import LiteralValue
from sqlparse.tokens import Error, Name, Punctuation

from metrics_layer.core.model.base import MetricsLayerBase
from metrics_layer.core.model.field import Field as MetricsLayerField
from metrics_layer.core.model.filter import MetricsLayerFilterExpressionType
from metrics_layer.core.sql.pypika_types import LiteralValueCriterion
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.core.sql.query_errors import ParseError


class MetricsLayerFilter(MetricsLayerBase):
    """
    An internal representation of a Filter (WHERE or HAVING clause)
    defined in a MetricsLayerQuery.

    definition: {"field", "expression", "value"}
    """

    def __init__(
        self, definition: Dict = {}, design: MetricsLayerDesign = None, filter_type: str = None
    ) -> None:
        definition = deepcopy(definition)

        # The design is used for filters in queries against specific designs
        #  to validate that all the tables and attributes (columns/aggregates)
        #  are properly defined in the design
        self.design = design
        self.is_literal_filter = "literal" in definition
        self.query_type = self.design.query_type
        self.filter_type = filter_type

        self.validate(definition)

        if not self.is_literal_filter:
            self.expression_type = MetricsLayerFilterExpressionType.parse(definition["expression"])

        super().__init__(definition)

    def validate(self, definition: Dict) -> None:
        """
        Validate the Filter definition
        """
        key = definition.get("field", None)
        filter_literal = definition.get("literal", None)

        if key is None and filter_literal is None:
            raise ParseError(f"An attribute key or literal was not provided for filter '{definition}'.")

        if key is None and filter_literal:
            return

        if definition["expression"] == "UNKNOWN":
            raise NotImplementedError(f"Unknown filter expression: {definition['expression']}.")

        if (
            definition.get("value", None) is None
            and definition["expression"] != "is_null"
            and definition["expression"] != "is_not_null"
        ):
            raise ParseError(f"Filter expression: {definition['expression']} needs a non-empty value.")

        if self.design:
            # Will raise ParseError if not found
            try:
                self.field = self.design.get_field(key)
            except ParseError:
                raise ParseError(
                    f"We could not find field {self.field_name} in explore {self.design.explore.name}"
                )

            if self.design.query_type == "BIGQUERY" and isinstance(definition["value"], datetime.datetime):
                cast_func = "DATETIME" if self.field.datatype == "date" else "TIMESTAMP"
                definition["value"] = LiteralValue(f"{cast_func}('{definition['value']}')")

            if self.field.type == "yesno" and "False" in definition["value"]:
                definition["expression"] = "boolean_false"

            if self.field.type == "yesno" and "True" in definition["value"]:
                definition["expression"] = "boolean_true"

    def sql_query(self):
        if self.is_literal_filter:

            return LiteralValueCriterion(self.replace_fields_literal_filter())
        return self.criterion(LiteralValue(self.field.sql_query(self.query_type, self.design.base_view_name)))

    def replace_fields_literal_filter(self):
        tokens = self._parse_sql_literal(self.literal)

        if self.filter_type == "where":
            extra_args = {"field_type": None}
        else:
            extra_args = {"field_type": "measure", "type": "number"}
        view = self.design.get_view(self.design.base_view_name)
        field = MetricsLayerField({"sql": "".join(tokens), "name": None, **extra_args}, view=view)
        return field.sql_query(self.query_type, view.name)

    def _parse_sql_literal(self, clause: str):
        generator = list(sqlparse.parse(clause)[0].flatten())
        tokens, field_names = [], []
        for i, token in enumerate(generator):
            not_already_added = i == 0 or str(generator[i - 1]) != "."

            if token.ttype == Name and not_already_added:
                try:
                    field = self.design.get_field(str(token))
                    tokens.append("${" + field.view.name + "." + str(token) + "}")
                except Exception:
                    field_names.append(str(token))
            elif token.ttype == Name and not not_already_added:
                pass
            elif token.ttype == Punctuation and str(token) == ".":
                if generator[i - 1].ttype == Name and generator[i + 1].ttype == Name:
                    field = self.design.get_field(f"{field_names[-1]}.{str(generator[i+1])}")
                    tokens.append("${" + field.view.name + "." + str(generator[i + 1]) + "}")
            elif token.ttype != Error:
                tokens.append(str(token))

        return tokens

    def criterion(self, field: Field) -> Criterion:
        """
        Generate the Pypika Criterion for this filter

        field: Pypika Field as we do not know the base table this filter is
                evaluated against (is it the base table or an intermediary table?)

        We have to use the following cases as PyPika does not allow an str
         representation of the clause on its where() and having() functions
        """
        criterion_strategies = {
            MetricsLayerFilterExpressionType.LessThan: lambda f: f < self.value,
            MetricsLayerFilterExpressionType.LessOrEqualThan: lambda f: f <= self.value,
            MetricsLayerFilterExpressionType.EqualTo: lambda f: f == self.value,
            MetricsLayerFilterExpressionType.NotEqualTo: lambda f: f != self.value,
            MetricsLayerFilterExpressionType.GreaterOrEqualThan: lambda f: f >= self.value,
            MetricsLayerFilterExpressionType.GreaterThan: lambda f: f > self.value,
            MetricsLayerFilterExpressionType.Like: lambda f: f.like(self.value),
            MetricsLayerFilterExpressionType.Contains: lambda f: f.like(f"%{self.value}%"),
            MetricsLayerFilterExpressionType.DoesNotContain: lambda f: f.not_like(f"%{self.value}%"),
            MetricsLayerFilterExpressionType.ContainsCaseInsensitive: lambda f: f.ilike(f"%{self.value}%"),
            MetricsLayerFilterExpressionType.DoesNotContainCaseInsensitive: lambda f: f.not_ilike(
                f"%{self.value}%"
            ),
            MetricsLayerFilterExpressionType.StartsWith: lambda f: f.like(f"{self.value}%"),
            MetricsLayerFilterExpressionType.EndsWith: lambda f: f.like(f"%{self.value}"),
            MetricsLayerFilterExpressionType.DoesNotStartWith: lambda f: f.not_like(f"{self.value}%"),
            MetricsLayerFilterExpressionType.DoesNotEndWith: lambda f: f.not_like(f"%{self.value}"),
            MetricsLayerFilterExpressionType.StartsWithCaseInsensitive: lambda f: f.ilike(f"{self.value}%"),
            MetricsLayerFilterExpressionType.EndsWithCaseInsensitive: lambda f: f.ilike(f"%{self.value}"),
            MetricsLayerFilterExpressionType.DoesNotStartWithCaseInsensitive: lambda f: f.not_ilike(
                f"{self.value}%"
            ),
            MetricsLayerFilterExpressionType.DoesNotEndWithCaseInsensitive: lambda f: f.not_ilike(
                f"%{self.value}"
            ),
            MetricsLayerFilterExpressionType.IsNull: lambda f: f.isnull(),
            MetricsLayerFilterExpressionType.IsNotNull: lambda f: f.notnull(),
            MetricsLayerFilterExpressionType.IsIn: lambda f: f.isin(self.value),
            MetricsLayerFilterExpressionType.IsNotIn: lambda f: f.isin(self.value).negate(),
            MetricsLayerFilterExpressionType.BooleanTrue: lambda f: f,
            MetricsLayerFilterExpressionType.BooleanFalse: lambda f: f.negate(),
        }

        try:
            return criterion_strategies[self.expression_type](field)
        except KeyError:
            raise NotImplementedError(f"Unknown filter expression_type: {self.expression_type}.")
